package com.tzld.piaoquan.ad.engine.service.score.deprecated;

import com.tzld.piaoquan.ad.engine.commons.score.ScoreParam;
import com.tzld.piaoquan.ad.engine.commons.util.NumUtil;
import com.tzld.piaoquan.ad.engine.service.score.RankService680;
import com.tzld.piaoquan.ad.engine.service.score.dto.AdDirectionScore;
import com.tzld.piaoquan.ad.engine.service.score.dto.AdPlatformCreativeDTO;
import com.tzld.piaoquan.ad.engine.service.score.param.RankRecommendRequestParam;
import com.tzld.piaoquan.recommend.feature.domain.ad.base.AdRankItem;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.MapUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.*;
import java.util.stream.Collectors;

@Slf4j
@Service
public class TacticsAndLRModelScoreRankService {

    @Autowired
    private RankService680 rankService680;

    public List<AdRankItem> adItemRank(RankRecommendRequestParam requestParam, ScoreParam scoreParam) {

        // LR模型打分结果
        List<AdRankItem> result = rankService680.adItemRank(requestParam, scoreParam);

        // List<JSONObject> collect = result.stream().map(item -> {
        //     JSONObject json = new JSONObject();
        //     json.put("cid", item.getAdId());
        //     json.put("score", item.getScore());
        //     json.put("modelScore", item.getLrScore());
        //     json.put("allfeature", item.getFeatureMap());
        //     return json;
        // }).collect(Collectors.toList());

        Map<Long, AdDirectionScore> adDirectionScoreMap = requestParam.getAdIdList().stream()
                .collect(Collectors.toMap(AdPlatformCreativeDTO::getCreativeId, AdPlatformCreativeDTO::getAdDirectionScore));

        for (AdRankItem adRankItem : result) {

            this.calcDirectionScore(adRankItem, adDirectionScoreMap.get(adRankItem.getAdId()));

            double s1 = adRankItem.getAdDirectionScore();
            // 模型的打分结果已经乘CPA，此处不需要重复乘CPA
            double s2 = adRankItem.getScore();
            adRankItem.setScore(s1 * s2);

        }

        Collections.sort(result);
        return result;
    }

    private void calcDirectionScore(AdRankItem adRankItem, AdDirectionScore adDirectionScore) {
        if (Objects.isNull(adDirectionScore) || MapUtils.isEmpty(adDirectionScore.getScoreDetail())) {
            adRankItem.getScoreMap().put("adDirectionScore", 1.0);
            adRankItem.setAdDirectionScore(1);
            return;
        }
        double exponent = Objects.nonNull(adDirectionScore.getExponent()) ? adDirectionScore.getExponent() : 1;
        Map<String, String> scoreDetail = adDirectionScore.getScoreDetail();

        double network = Double.parseDouble(scoreDetail.getOrDefault("network", "1"));
        double phoneModel = Double.parseDouble(scoreDetail.getOrDefault("phoneModel", "1"));
        double area = Double.parseDouble(scoreDetail.getOrDefault("area", "1"));
        double peoplePackage = Double.parseDouble(scoreDetail.getOrDefault("peoplePackage", "1"));


        double excludeMin = 1.0;

        if (scoreDetail.containsKey("excludeConv") || scoreDetail.containsKey("excludeClick") || scoreDetail.containsKey("excludeView")) {
            double excludeConv = Double.parseDouble(scoreDetail.getOrDefault("excludeConv", "10000"));
            double excludeClick = Double.parseDouble(scoreDetail.getOrDefault("excludeClick", "10000"));
            double excludeView = Double.parseDouble(scoreDetail.getOrDefault("excludeView", "10000"));
            excludeMin = NumUtil.min(excludeConv, excludeClick, excludeView);
        }

        double s1 = (network * phoneModel * area * peoplePackage * excludeMin);
        double s2 = Math.pow(s1, exponent);
        adRankItem.getScoreMap().put("adDirectionScore", s2);
        adRankItem.setAdDirectionScore(s2);

        Map<String, String> scoreDetailMap = new HashMap<>(scoreDetail);
        scoreDetailMap.put("exponent", String.valueOf(exponent));
        scoreDetailMap.put("excludeMin", String.valueOf(excludeMin));
        adRankItem.getFeatureMap().putAll(scoreDetailMap);
    }

}
